#include <windows.h>
#include <stdio.h>

void print_help() {
    printf("Usage:\n"
           "  screcover [-<switches>] <source file> <target file>\n"
           "Switches:\n"
           "  -b <blocksize>      sets physical block size for media. 2048 is default.\n"
           "  -o[b|k|m] <offset>  offset in bytes (or optionally in blocks, KBytes or MBytes)\n"
           "                      to start recovery from\n"
           "  -c[b|k|m] <count>   number of bytes (or optionally in blocks, KBytes or MBytes)\n"
           "                      to recover\n"
           "  -r <retry>          number of retry attempts. 3 is default\n"
           "  -l <logfile>        log unreliable and bad blocks to <logfile>\n"
           "  -?                  display this help\n"
           "Note:\n"
           "  1 KByte = 1024 bytes\n"
           "  1 MByte = 1024 KBytes = 1024*1024 bytes\n"
           "  byte offset or byte count is specified, it will be rounded to blocksize\n"
           "Examples:\n"
           "  screcover -o 5000000 X:\\Video.dat E:\\Temp\\Restored_Film.dat\n"
           "    will copy data starting from byte offset 5000000 to end of file"
           "  screcover -om 1300 -cm 1000 X:\\Video.dat E:\\Temp\\Restored_Film.dat\n"
           "    will copy not more than 1000Mb of data starting from byte offset 1300Mb"
          );
    exit(0);
}

__int64
get_file_size(
    HANDLE h
    )
{
    LONG hsz = 0;
    LONG lsz;

    lsz = SetFilePointer(h, 0, &hsz, FILE_END);
    return (((__int64)hsz) << 32) | lsz;
}

__int64
set_file_pointer(
    HANDLE h,
    __int64 sz
    )
{
    ULONG hsz = (ULONG)(sz >> 32);
    ULONG lsz = (ULONG)sz;

    lsz = SetFilePointer(h, lsz, (PLONG)&hsz, FILE_BEGIN);
    return (((__int64)hsz) << 32) | lsz;
}

int
set_file_size(
    HANDLE h,
    __int64 sz
    )
{
    set_file_pointer(h, sz);
    return SetEndOfFile(h);
}

void main (void) {
    int i, j;
    WCHAR a;
    ULONG bs = 2048;
    ULONG retry = 3;
    ULONG r;
    PWCHAR in_name = NULL;
    PWCHAR out_name = NULL;
    PWCHAR log_name = NULL;
    int break_sw;
    int block_offs = 0;
    int block_counter = 0;
    __int64 offs = 0;
    __int64 counter = 0;
    __int64 f_size;
    __int64 f_ptr;
    __int64 bads = 0;
    __int64 susp = 0;
    __int64 copy_count = 0;
    char* buff;
    ULONG read_bytes;
    BOOLEAN ok;
    __int64 p;

    WCHAR*  CmdLine;
    PWCHAR* argv;
    int argc;

    CHAR LogBuffer[128];

    HANDLE in_h;
    HANDLE out_h;
    HANDLE log_h = NULL;

    CmdLine = GetCommandLineW();
    argv = CommandLineToArgvW(CmdLine, &argc);

    printf("Scratched File Copy utility v1.2 (c) by Alter, 2004\n"
           "Home site http://www.alter.org.ua\n");

    if(argc < 3)
        print_help();
    for(i=1; i<argc; i++) {
        if(!argv[i])
            continue;
        if(argv[i][0] != '-') {
            if(!in_name) {
                in_name = argv[i];
                continue;
            } else
            if(!out_name) {
                out_name = argv[i];
                continue;
            } else {
                print_help();
            }
        }
        j=1;
        break_sw = 0;
        while(!break_sw &&
              (a = argv[i][j]) && (a != ' ') && (a != '\t')) {
            switch(a) {
            case 'b' :
                i++;
                if(!argv[i])
                    print_help();
                swscanf(argv[i], L"%d", &bs);
                if(!bs)
                    bs = 2048;
                break_sw = 1;
                break;
            case 'l' :
                i++;
                if(!argv[i])
                    print_help();
                log_name = argv[i];
                break_sw = 1;
                break;
            case 'o' :
                a = argv[i][j+1];
                switch(a) {
                case 'b':
                    i++;
                    if(!argv[i])
                        print_help();
                    swscanf(argv[i], L"%I64d", &offs);
                    block_offs = TRUE;
                case 'k':
                    i++;
                    if(!argv[i])
                        print_help();
                    swscanf(argv[i], L"%I64d", &offs);
                    offs *= 1024;
                case 'm':
                    i++;
                    if(!argv[i])
                        print_help();
                    swscanf(argv[i], L"%I64d", &offs);
                    offs *= 1024*1024;
                default:
                    i++;
                    if(!argv[i])
                        print_help();
                    swscanf(argv[i], L"%I64d", &offs);
                }
                break_sw = 1;
                break;
            case 'c' :
                a = argv[i][j+1];
                switch(a) {
                case 'b':
                    i++;
                    if(!argv[i])
                        print_help();
                    swscanf(argv[i], L"%I64d", &counter);
                    block_counter = TRUE;
                case 'k':
                    i++;
                    if(!argv[i])
                        print_help();
                    swscanf(argv[i], L"%I64d", &counter);
                    counter *= 1024;
                case 'm':
                    i++;
                    if(!argv[i])
                        print_help();
                    swscanf(argv[i], L"%I64d", &counter);
                    counter *= 1024*1024;
                default:
                    i++;
                    if(!argv[i])
                        print_help();
                    swscanf(argv[i], L"%I64d", &counter);
                }
                break_sw = 1;
                break;
            case 'r' :
                i++;
                if(!argv[i])
                    print_help();
                swscanf(argv[i], L"%d", &retry);
                break_sw = 1;
                break;
            case '?' :
            default:
                print_help();
            }
            j++;
        }
    }
    if(!in_name || !out_name)
        print_help();

    in_h = CreateFileW(in_name, GENERIC_READ, FILE_SHARE_READ, NULL,
                      OPEN_EXISTING,
                      FILE_ATTRIBUTE_NORMAL /*| FILE_FLAG_NO_BUFFERING*/ | FILE_FLAG_SEQUENTIAL_SCAN,
                      NULL);
    if(in_h == (HANDLE)(-1)) {
        wprintf(L"Can't open source file %s\n"
               L"exiting...\n", in_name);
        exit (-1);
    }

    buff = (char*)GlobalAlloc(GMEM_DISCARDABLE, bs);
    if(!buff) {
        printf("Can't alloc %d bytes for temporary buffer\n"
               "exiting...\n", bs);
        exit (-2);
    }

    out_h = CreateFileW(out_name, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ,
                      NULL,
                      OPEN_ALWAYS,
                      FILE_ATTRIBUTE_NORMAL /*| FILE_FLAG_NO_BUFFERING*//* | FILE_FLAG_SEQUENTIAL_SCAN*/,
                      NULL);
    if(out_h == (HANDLE)(-1)) {
        read_bytes = GetLastError();
        wprintf(L"Can't create target file %s\n"
               L"exiting...\n", out_name);
        exit (-3);
    }

    if(log_name) {
        log_h = CreateFileW(log_name, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ,
                          NULL,
                          CREATE_ALWAYS,
                          FILE_ATTRIBUTE_NORMAL /*| FILE_FLAG_NO_BUFFERING*//* | FILE_FLAG_SEQUENTIAL_SCAN*/,
                          NULL);
        if(log_h == (HANDLE)(-1)) {
            read_bytes = GetLastError();
            wprintf(L"Can't create log file %s\n"
                   L"exiting...\n", log_name);
            exit (-6);
        }
    }

    f_size = get_file_size(in_h);
    if(!f_size) {
        wprintf(L"Empty source file %s\n"
               L"exiting...\n", out_name);
        exit (-4);
    }

    if(!set_file_size(out_h, f_size)) {
        wprintf(L"Can't set size of target file %s to %I64d bytes\n"
               L"exiting...\n", out_name, f_size);
        exit (-5);
    }

    // align offs and counter
    offs = offs & ~((unsigned __int64)(bs-1));
    counter = (counter + bs - 1) & ~((unsigned __int64)(bs-1));

    f_ptr = offs;
    if(block_offs) {
        f_ptr *= bs;
    }
    if(block_counter) {
        counter *= bs;
    }

    wprintf(L"%s -> %s\n%I64d bytes\n",
           in_name, out_name, f_size);
    wprintf(L"Starting from offset %I64d bytes\n",
           offs);
    if(counter) {
        wprintf(L"Recovery range %I64d bytes\n",
               counter);
    }

    p = f_size/50;
    if(!p) p = 1;

    while(f_ptr <= f_size-bs) {
        if(counter && copy_count >= counter) {
            printf("Copied %I64d bytes, %I64d unreliable, %I64d in bad blocks\n", copy_count, susp, bads);
            exit(0);
        }
        if((f_ptr % p) < bs)
            printf(".");
        memset(buff, 0, bs);
        ok = FALSE;
        for(r=0; r<retry; r++) {
            set_file_pointer(in_h, f_ptr);
            set_file_pointer(out_h, f_ptr);
            read_bytes = 0;
            if(!ReadFile(in_h, buff, bs, &read_bytes, NULL) ||
                read_bytes != bs) {
                read_bytes = GetLastError();
                if(!r) {
                    printf("Can't read block at offset %I64d\n", f_ptr);
                }
            } else {
                if(r) {
                    printf("Read on attempt #%d\n", r+1);
                }
                WriteFile(out_h, buff, bs, &read_bytes, NULL);
                if(read_bytes != bs) {
                    printf("Can't write block at offset %I64d\n", f_ptr);
                } else {
                    ok = TRUE;
                }
                break;
            }
        }
        if(!ok) {
            bads += bs;
            if(log_h) {
                sprintf(LogBuffer, "bad at %I64d size %d\n", f_ptr, bs);
                WriteFile(log_h, LogBuffer, strlen(LogBuffer), &read_bytes, NULL);
            }
        } else
        if(r) {
            susp += bs;
            if(log_h) {
                sprintf(LogBuffer, "unreliable at %I64d size %d\n", f_ptr, bs);
                WriteFile(log_h, LogBuffer, strlen(LogBuffer), &read_bytes, NULL);
            }
        }
        f_ptr += bs;
        copy_count += bs;
    }

    printf("\n");
    bs = (ULONG)(min(bs, f_size - f_ptr));
    if(!bs || (counter && copy_count >= counter)) {
        printf("Copied %I64d bytes, %I64d unreliable, %I64d in bad blocks\n", copy_count, susp, bads);
        exit(0);
    }
    memset(buff, 0, bs);
    ReadFile(in_h, buff, bs, &read_bytes, NULL);
    if(read_bytes != bs) {
        printf("Can't read block at offset %I64d\n", f_ptr);
        bads += bs;
    }
    WriteFile(out_h, buff, bs, &read_bytes, NULL);
    if(read_bytes != bs) {
        printf("Can't write block at offset %I64d\n", f_ptr);
    }

    printf("Copied %I64d bytes, %I64d unreliable, %I64d in bad blocks\n", copy_count, susp, bads);
    exit(0);
}
